/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2007 Bull S.A.S.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: Activator.java 5988 2011-09-28 19:14:39Z cazauxj $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.deploy.impl.osgi;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;
import org.ow2.util.ee.deploy.api.helper.IDeployableHelper;
import org.ow2.util.archive.api.IArchiveManager;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.ee.deploy.api.deployer.IDeployer;
import org.ow2.util.ee.deploy.api.deployer.IDeployerManager;
import org.ow2.util.ee.deploy.api.deployer.IDeployerManagerCallback;
import org.ow2.util.ee.deploy.impl.deployer.DeployerManager;
import org.ow2.util.ee.deploy.impl.helper.DeployableAnalyserTracker;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperComponent;

/**
 * Activator which registers DeployerManager, ArchiveManager and OSGi Deployer
 * services.
 * @author Francois Fornaciari
 */
public class Activator implements BundleActivator, ServiceListener {

    /**
     * The bundle context.
     */
    private BundleContext bundleContext = null;

    /**
     * The deployer manager.
     */
    private DeployerManager deployerManager = null;

    /**
     * The DeployableAnalyser tracker
     */
    private DeployableAnalyserTracker deployableAnalyserTracker = null;

    /**
     * Called when this bundle is started so the Framework can perform the
     * bundle-specific activities necessary to start this bundle. This method
     * can be used to register services or to allocate any resources that this
     * bundle needs.
     * @param context The execution context of the bundle being started.
     * @throws Exception If the start fails
     */
    public void start(final BundleContext bundleContext) throws Exception {
        this.bundleContext = bundleContext;

        // Create the OSGi deployer
        OSGiDeployer osgiDeployer = new OSGiDeployer(bundleContext);

        // Create a new deployer manager
        this.deployerManager = new DeployerManager();

        // Register the deployer on the deployer manager
        this.deployerManager.register(osgiDeployer);

        // Create a new DeployableHelperComponent
        this.deployableAnalyserTracker = new DeployableAnalyserTracker(bundleContext);
        this.deployableAnalyserTracker.start();  //start the DeployableAnalyser tracker
        DeployableHelperComponent deployableHelperComponent = new DeployableHelperComponent(this.deployableAnalyserTracker);

        // Register the deployer manager and the deployer and the DeployerHelperComponent as an OSGi service
        bundleContext.registerService(IDeployerManager.class.getName(), this.deployerManager, null);
        bundleContext.registerService(IDeployer.class.getName(), osgiDeployer, null);
        bundleContext.registerService(IDeployableHelper.class.getName(), deployableHelperComponent, null);

        // Add also the archive manager
        ArchiveManager archiveManager = ArchiveManager.getInstance();
        bundleContext.registerService(IArchiveManager.class.getName(), archiveManager, null);

        // Add existing DeployerManager callbacks registered as OSGi services
        addCallbacks();
        
        // Handle callback registrations/unregistrations
        bundleContext.addServiceListener(this, "(objectClass=" + IDeployerManagerCallback.class.getName() + ")");
    }

    /**
     * Called when this bundle is stopped so the Framework can perform the
     * bundle-specific activities necessary to stop the bundle.
     * @param context The execution context of the bundle being stopped.
     */
    public void stop(final BundleContext bc) {
        //stop the DeployableAnalyser tracker
        this.deployableAnalyserTracker.stop();
    }

    /**
     * {@inheritDoc}
     */
    public void serviceChanged(ServiceEvent event) {
        ServiceReference serviceRef = event.getServiceReference();
        IDeployerManagerCallback callback = (IDeployerManagerCallback) this.bundleContext.getService(serviceRef);
        switch (event.getType()) {
        case ServiceEvent.REGISTERED:
            deployerManager.addCallback(callback);
            break;
        case ServiceEvent.UNREGISTERING:
            deployerManager.removeCallback(callback);
            break;
        default:
            break;
        }
    }

    /**
     * Add a new {@IDeplyerManagerCallback} registered as an OSGi service.
     * @throws InvalidSyntaxException If the service reference cannot be retrieved
     */
    private void addCallbacks() throws InvalidSyntaxException {
        ServiceReference[] serviceReferences = this.bundleContext.getServiceReferences(
                IDeployerManagerCallback.class.getName(), null);
        if (serviceReferences != null) {
            for (ServiceReference serviceReference : serviceReferences) {
                IDeployerManagerCallback callback = (IDeployerManagerCallback) bundleContext.getService(serviceReference);
                deployerManager.addCallback(callback);
            }
        }
    }

}
