/**
 * OW2 Util
 * Copyright (C) 2008-2011 Bull S.A.S.
 * Contact: easybeans@ow2.org
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
* USA
*
* --------------------------------------------------------------------------
* $Id: OSGiDeployer.java 6068 2011-12-24 12:55:55Z cazauxj $
* --------------------------------------------------------------------------
*/

package org.ow2.util.ee.deploy.impl.osgi;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceReference;
import org.osgi.service.startlevel.StartLevel;
import org.ow2.util.archive.api.IArchiveMetadata;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.OSGiDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.ee.deploy.impl.deployer.AbsDeployer;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * OSGi deployer that is able to deploy an OSGi Deployable
 * @author Florent BENOIT
 * @author Jeremy Cazaux (support to the deployment of a list of deployables)
 */
public class OSGiDeployer extends AbsDeployer<OSGiDeployable> {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(OSGiDeployer.class);

    /**
     * Bundle Context to use for the deployment.
     */
    private BundleContext bundleContext = null;

    /**
     * List of Bundles that have been installed.
     */
    private Map<String, Bundle> installedBundles = null;

    /**
     * OSGi StartLevel service reference.
     */
    private StartLevel startLevel = null;

    /**
     * Bundle symbolic name property
      */
    public static String BUNDLE_SYMBOLIC_NAME = "Bundle-SymbolicName";

    /**
     * Bundle version property
     */
    public static String BUNDLE_VERSION = "Bundle-Version";

    /**
     * Default constructor.
     * @param bundleContext OSGi bundle context
     */
    public OSGiDeployer(final BundleContext bundleContext) {
        this.bundleContext = bundleContext;
        this.installedBundles = new HashMap<String, Bundle>();

        // Get the StartLevel service reference
        ServiceReference serviceReference = bundleContext.getServiceReference(StartLevel.class.getName());
        setStartLevel((StartLevel) bundleContext.getService(serviceReference));
    }

    /**
     * @param service the StartLevel service
     */
    public void setStartLevel(StartLevel service) {
        this.startLevel = service;
    }

    /**
     * Undeploy the given Bundle.
     * @param osgiDeployable the deployable to remove.
     * @throws DeployerException if the bundle is not undeployed.
     */
    public void doUndeploy(final IDeployable<OSGiDeployable> osgiDeployable) throws DeployerException {
        logger.debug("Undeploying {0}", osgiDeployable);

        //stop the bundle
        Bundle bundle = stopBundle(osgiDeployable);

        if (bundle != null) {
            //uninstall the bundle
            uninstallBundle(osgiDeployable, bundle);
        }
    }


    /**
     * Deploy the given Bundle.
     * @param deployable the deployable to add.
     * @throws DeployerException if the Bundle is not deployed.
     */
    public void doDeploy(final IDeployable<OSGiDeployable> deployable) throws DeployerException {
        logger.debug("Deploying {0}", deployable);

        // Cast the deployable only once
        OSGiDeployable osgiDeployable = OSGiDeployable.class.cast(deployable);

        // First, install the bundle
        Bundle bundle = installBundle(osgiDeployable);

        if (bundle != null) {
            // And then start it
            startBundle(osgiDeployable, bundle);

            // Add it as installed
            addInstalledBundle(osgiDeployable, bundle);
        }
    }

    /**
     * Checks if the given deployable is supported by the Deployer.
     * @param deployable the deployable to be checked
     * @return true if it is supported, else false.
     */
    public boolean supports(final IDeployable<?> deployable) {
        return OSGiDeployable.class.isInstance(deployable);
    }

    /**
     * Check if the given deployable is deployable or not.
     * @param deployable the deployable to check.
     * @throws DeployerException if the deployable is not supported.
     */
    protected void check(final IDeployable<?> deployable) throws DeployerException {
        if (bundleContext == null) {
            throw new DeployerException("The deployment of the deployable'" + deployable
                    + "' is not available as there is no bundle context set on this deployer.");
        }

        super.check(deployable);
    }

    /**
     * Checks if the given deployable is deployed or not.
     * @param deployable test if a given deployable is already deployed.
     * @return true if the deployable is deployed.
     * @throws DeployerException if the undeploy operation fails.
     */
    @Override
    public boolean isDeployed(final IDeployable<OSGiDeployable> deployable) throws DeployerException {

        List<Bundle> bundles = Arrays.asList(this.bundleContext.getBundles());

        IArchiveMetadata archiveMetadata = deployable.getArchive().getMetadata();

        String symbolicName = archiveMetadata.get(BUNDLE_SYMBOLIC_NAME);
        String version = archiveMetadata.get(BUNDLE_VERSION);

        for (Bundle bundle: bundles) {

            if (bundle.getSymbolicName().equals(symbolicName) && bundle.getVersion().toString().equals(version)) {

                //if the state of the bundle is "Installed" or "Resolved", we start the bundle
                if ((bundle.getState() == Bundle.INSTALLED || bundle.getState() == Bundle.RESOLVED)) {
                    startBundle(OSGiDeployable.class.cast(deployable), bundle);
                }

                this.installedBundles.put(getURL(deployable).toExternalForm(), bundle);

                return true;
            }

        }

        return super.isDeployed(deployable);
    }

    /**
     * Deploy the list of given Bundle.
     * @param deployables the list of deployable to add.
     */
    @Override
    public List<IDeployable<OSGiDeployable>> doDeploy(List<IDeployable<OSGiDeployable>> deployables) {

        //list of deployable which have a failed deployment
        List<IDeployable<OSGiDeployable>> failed = new ArrayList<IDeployable<OSGiDeployable>>();

        Map<OSGiDeployable, Bundle> osgiDeployableBundleMap = new HashMap<OSGiDeployable, Bundle>();

        //first, install all bundles
        for (IDeployable deployable: deployables) {
            this.logger.debug("Deploying {0}", deployable);

            // Cast the deployable only once
            OSGiDeployable osgiDeployable = OSGiDeployable.class.cast(deployable);

            //install the bundle
            Bundle bundle;
            try {
                bundle = installBundle(osgiDeployable);
                if (bundle != null) {
                    //Keep the link OSGiDeployable -> Bundle
                    osgiDeployableBundleMap.put(osgiDeployable, bundle);
                }
            } catch (DeployerException e) {
                logger.error("Bundle {0} cannot be installed", osgiDeployable, e);
                failed.add(osgiDeployable);
            }
        }

        //And then start all bundles
        for (Map.Entry<OSGiDeployable, Bundle> entry : osgiDeployableBundleMap.entrySet()) {

            OSGiDeployable osgiDeployable = entry.getKey();
            Bundle bundle = entry.getValue();

            //start the bundle
            startBundle(osgiDeployable, bundle);

            // Add it as installed
            try {
                addInstalledBundle(osgiDeployable, bundle);
            } catch (DeployerException e) {
                logger.error("Cannot add {0} to the list of installed bundle", osgiDeployable, e);
                failed.add(osgiDeployable);
            }

        }
        return failed;
    }

    /**
     * Undeploy a list of Bundle without dependencies problem.
     * @param osgiDeployables The list of deployable to remove.
     */
    @Override
    public List<IDeployable<OSGiDeployable>> doUndeploy(final List<IDeployable<OSGiDeployable>> osgiDeployables) {

        //list of deployable which have a failed undeployment
        List<IDeployable<OSGiDeployable>> failed = new ArrayList<IDeployable<OSGiDeployable>>();

        Map<IDeployable<OSGiDeployable>, Bundle> osgiDeployableBundleMap = new HashMap<IDeployable<OSGiDeployable>, Bundle>();

        for (IDeployable osgiDeployable: osgiDeployables) {

            this.logger.debug("Undeploying {0}", osgiDeployable);

            Bundle bundle;
            try {
                bundle = stopBundle(osgiDeployable);
                //keep the link deployable -> bundle if the bundle is not already uninstalled
                if (bundle != null) {
                    osgiDeployableBundleMap.put(osgiDeployable, bundle);
                }
            } catch (DeployerException e) {
                logger.error("Cannot stop bundle {0}", osgiDeployable, e);
                failed.add(osgiDeployable);
            }
        }

        //And then uninstall all installed bundles
        for (Map.Entry<IDeployable<OSGiDeployable>, Bundle> entry : osgiDeployableBundleMap.entrySet()) {

            IDeployable<OSGiDeployable> osgiDeployable = entry.getKey();

            try {
                uninstallBundle(osgiDeployable, entry.getValue());
            } catch (DeployerException e) {
                logger.error("Cannot uninstall bundle {0}", osgiDeployable, e);
                failed.add(osgiDeployable);
            }
        }
        return failed;
    }


    /**
     * Install an OSGi deployable
     * @param osgiDeployable The OSGi deployable to install
     * @return the installed bundle if the OSGi deployable has been installed. Return null if the bundle is already
     * installed.
     * @throws DeployerException
     */
    protected Bundle installBundle(OSGiDeployable osgiDeployable) throws DeployerException {
        Bundle osgiBundle;
        try {
            String url = getURL(osgiDeployable).toExternalForm();
            if (osgiDeployable.isReference()) {
                url = "reference:".concat(url);
            }

            //install the bundle
            osgiBundle = this.bundleContext.installBundle(url);

            // Set the initial bundle start level
            startLevel.setInitialBundleStartLevel(osgiDeployable.getStartLevel());

        } catch (BundleException e) {
            // Check the bundle exception type
            if (e.getType() == BundleException.DUPLICATE_BUNDLE_ERROR) {
                logger.warn("Cannot deploy a bundle that is already installed", e.getMessage());
                return null;
            } else {
                throw new DeployerException("Cannot install the bundle '" + osgiDeployable + "'.", e);
            }
        }

        return osgiBundle;
    }

    /**
     * @param osgiDeployable The OSGi deployable
     * @param bundle Its associated Bundle to start
     */
    protected void startBundle(OSGiDeployable osgiDeployable, Bundle bundle) {

        if (osgiDeployable.isStart()) {

            // And then start it
            try {
                if (osgiDeployable.isStartTransient()) {
                    bundle.start(Bundle.START_TRANSIENT);
                } else {
                    bundle.start();
                }
            } catch (BundleException e) {
                logger.error("Cannot start the bundle ''{0}''.", osgiDeployable, e);
            }
        }
    }

    /**
     * @param osgiDeployable The OSGi deployable
     * @param bundle Its associated Bundle
     * @throws DeployerException
     */
    private void addInstalledBundle(OSGiDeployable osgiDeployable, Bundle bundle) throws DeployerException {
        this.installedBundles.put(getURL(osgiDeployable).toExternalForm(), bundle);
    }

    /**
     *
     * @param osgiDeployable The OSGi deployable
     * @return its associated bundle if the OSGi deployable has been stop
     * @throws DeployerException is the deployable is already uninstalled
     */
    protected Bundle stopBundle(IDeployable<OSGiDeployable> osgiDeployable) throws DeployerException {

        // Check if it has been installed ?
        if (!isDeployed(osgiDeployable)) {
            throw new DeployerException("Cannot undeploy the bundle '" + osgiDeployable + "' as it was not deployed.");
        }

        // get the bundle Object
        Bundle bundle = installedBundles.get(getURL(osgiDeployable).toExternalForm());

        // Check found
        if (bundle == null) {
            throw new DeployerException("No bundle found for the '" + osgiDeployable + "'.");
        }

        // Do not stop or uninstall if already uninstall/removed
        if (Bundle.UNINSTALLED != bundle.getState()) {
            // Stop the bundle
            try {
                bundle.stop();
            } catch (BundleException e) {
                throw new DeployerException("Cannot stop the bundle '" + osgiDeployable + "'.", e);
            }

            return bundle;

        } else {

            // Remove it of installed bundle
            this.installedBundles.remove(getURL(osgiDeployable).toExternalForm());

            return null;
        }
    }

    /**
     * Uninstalled a bundle
     * @param osgiDeployable The OSGi deployable
     * @param bundle Its associated bundle
     * @throws DeployerException if the bundle cannot be uninstalled
     */
    protected void uninstallBundle(IDeployable<OSGiDeployable> osgiDeployable, Bundle bundle) throws DeployerException {

        // Uninstall the bundle
        try {
            bundle.uninstall();
        } catch (BundleException e) {
            throw new DeployerException("Cannot uninstall the bundle '" + osgiDeployable + "'.", e);
        }

        // Remove it of installed bundle
        this.installedBundles.remove(getURL(osgiDeployable).toExternalForm());
    }
}
